#include <mex.h>
#include <cmath>
#include <algorithm>

#ifdef _MSC_VER
	#define __PARALLEL__
	#include <ppl.h>
#endif

typedef double im_type;
typedef unsigned int uint32;
typedef unsigned char uint8;


//
// prototype of functions
//
void box_filtering( size_t sy, size_t sx, im_type *J, const im_type *I, im_type *T, size_t ry, size_t rx );


//
// main IO
//
void mexFunction( int nlhs, mxArray* plhs[], int nrhs, const mxArray* prhs[] )
{
	if ( mxGetClassID( prhs[0] ) != mxDOUBLE_CLASS )
	{	mexErrMsgTxt( "Double class image is only supported." );
	}
	
	// input
	const im_type *I = (const im_type*)mxGetData( prhs[0] );
	size_t ndim = mxGetNumberOfDimensions( prhs[0] );
	const mwSize *dims = mxGetDimensions( prhs[0] );
	
	if ( ndim != 2 )
	{	mexErrMsgTxt( "Gray scale is only supported." );
	}	
	
	size_t sy, sx;
	sy = dims[0];
	sx = dims[1];
	
	// paramters
	size_t rx, ry;
	
	ry = (size_t)mxGetScalar( prhs[1] );
	rx = (size_t)mxGetScalar( prhs[2] );
	
	// output
	plhs[0] = mxCreateNumericArray( ndim, dims, mxDOUBLE_CLASS, mxREAL );
	im_type *J = (im_type*)mxGetData( plhs[0] );


	// temporal buff
	mxArray *mx_T = mxCreateNumericMatrix( sy, sx, mxDOUBLE_CLASS, mxREAL );
	im_type *T = (im_type*)mxGetData( mx_T );
	
	box_filtering( sy, sx, J, I, T, ry, rx );
	
	mxDestroyArray( mx_T );
}



//
// box filtering
//
inline void box_filtering_vdir( size_t sy, size_t sx, im_type *J, const im_type *I, im_type *T, size_t r );
inline void box_filtering_hdir( size_t sy, size_t sx, im_type *J, const im_type *I, im_type *T, size_t r );

void box_filtering( size_t sy, size_t sx, im_type *J, const im_type *I, im_type *T, size_t ry, size_t rx ) {
	box_filtering_vdir( sy, sx, J, I, T, ry );
	box_filtering_hdir( sy, sx, J, J, T, rx );
}





//
// Process in vertical direction
//
// #ifdef __PARALLEL__
//
// inline void box_filtering_vdir( size_t sy, size_t sx, im_type *J, const im_type *I, im_type *T, size_t r )
// {
// 	int w1( r+1 ), w2( sy-2*r-1 ), w3( r );
//
// 	Concurrency::parallel_for( size_t(0), size_t(sx), [&](size_t ix)
// 	{
// 		//
// 		// calculate cumulative sum
// 		//
// 		im_type cum_sum(0);
// 		size_t off(ix*sy);
//
// 		const im_type *pI(I+off), *pI_end;
// 		im_type *pJ(J+off), *pT(T+off), *pT1(pT), *pT2(pT+r);
//
//
// 		for ( pI_end=pI+sy;  pI<pI_end;  pI++, pT++ )
// 		{	*pT = (cum_sum += *pI);
// 		}
//
// 		//
// 		// calculate difference
// 		//
// 		im_type *pT1_end, *pT2_end;
//
// 		// top boundary
// 		pT2_end = pT2+w1;
//
// 		for ( ; pT2<pT2_end; pJ++, pT2++ )
// 		{	*pJ = *pT2;
// 		}
//
//
// 		// normal filtering
//  		pT1_end = pT1+w2;
//
// 		for ( ; pT1<pT1_end; pJ++, pT1++, pT2++ )
// 		{	*pJ = *pT2 - *pT1;
// 		}
//
// 		// bottom boundary
// 		pT1_end = pT1+w3;
// 		pT2--;
//
// 		for ( ; pT1<pT1_end; pJ++, pT1++ )
// 		{	*pJ = *pT2 - *pT1;
// 		}
// 	});
// }
//
// #else

inline void box_filtering_vdir( size_t sy, size_t sx, im_type *J, const im_type *I, im_type *T, size_t r ) {
	int w1( r+1 ), w2( sy-2*r-1 ), w3( r );
	
	for ( size_t ix = 0; ix < sx; ix++ ) {
		//
		// calculate cumulative sum
		//
		im_type cum_sum(0);
		int off(ix*sy);
		
		const im_type *pI(I+off), *pI_end;
		im_type *pJ(J+off), *pT(T+off), *pT1(pT), *pT2(pT+r);
		
		
		for ( pI_end=pI+sy;  pI<pI_end;  pI++, pT++ )
		{	cum_sum += *pI;
			*pT = cum_sum;
		}
		
		//
		// calculate difference
		//
		im_type *pT1_end, *pT2_end;
		
		// top boundary
		pT2_end = pT2+w1;
		
		for ( ; pT2<pT2_end; pJ++, pT2++ )
		{	*pJ = *pT2;
		}
		
		
		// normal filtering
		pT1_end = pT1+w2;
		
		for ( ; pT1<pT1_end; pJ++, pT1++, pT2++ )
		{	*pJ = *pT2 - *pT1;
		}
		
		// bottom boundary
		pT1_end = pT1+w3;
		pT2--;
		
		for ( ; pT1<pT1_end; pJ++, pT1++ )
		{	*pJ = *pT2 - *pT1;
		}
	}
}

// //
// // Old version 2011Mar14
// //
// // inline void box_filtering_vdir( size_t sy, size_t sx, im_type *pJ, const im_type *pI, im_type *pT, size_t r )
// // {
// // 	const im_type *pI_end;
// // 	im_type *pT1, *pT2;
// //
// // 	int w1( r+1 ), w2( sy-2*r-1 ), w3( r );
// // 	im_type *pT2_begin1(pT+r), *pT2_end1;
// // 	im_type *pT1_begin2(pT), *pT1_end2;
// // 	im_type *pT1_end3;
// //
// //
// // 	for ( size_t ix = 0; ix < sx; ix++ )
// // 	{
// // 		//
// // 		// calculate cumulative sum
// // 		//
// // 		im_type cum_sum(0);
// //
// // 		for ( pI_end=pI+sy;  pI<pI_end;  pI++, pT++ )
// // 		{	cum_sum += *pI;
// // 			*pT = cum_sum;
// // 		}
// //
// // 		//
// // 		// calculate difference
// // 		//
// // 		// top boundary
// // 		pT2 = pT2_begin1;  pT2_end1 = pT2+w1;
// //
// // 		for ( ; pT2<pT2_end1; pJ++, pT2++ )
// // 		{	*pJ = *pT2;
// // 		}
// // 		pT2_begin1 += sy;
// //
// // 		// normal filtering
// //  		pT1 = pT1_begin2;  pT1_end2 = pT1+w2;
// //
// // 		for ( ; pT1<pT1_end2; pJ++, pT1++, pT2++ )
// // 		{	*pJ = *pT2 - *pT1;
// // 		}
// // 		pT1_begin2 += sy;
// //
// // 		// bottom boundary
// // 		pT1_end3 = pT1+w3;
// // 		pT2--;
// //
// // 		for ( ; pT1<pT1_end3; pJ++, pT1++ )
// // 		{	*pJ = *pT2 - *pT1;
// // 		}
// // 	}
// // }
//
// #endif


//
// #ifdef __PARALLEL__
//
// //
// // Process in vertical direction
// //
// inline void box_filtering_hdir( size_t sy, size_t sx, im_type *J, const im_type *pI, im_type *T, size_t r )
// {
// 	im_type *pJ_begin, *pJ_end, *pT(T);
// 	im_type *pT1_begin, *pT2_begin;
//
// 	size_t w1, w2, w3;
//
// 	//
// 	// calculate cumulative sum
// 	//
// 	im_type *buff = new im_type [sy], *pbuff, *pbuff_end;
// 	std::fill( buff, buff+sy, 0 );
//
// 	for ( size_t ix = 0; ix < sx; ix++ )
// 	{
// 		for ( im_type *pbuff(buff), *pbuff_end(buff+sy);  pbuff<pbuff_end;  pbuff++, pI++, pT++ )
// 		{
// 			*pT = ( *pbuff += *pI );
// 		}
// 	}
//
// 	delete [] buff;
//
//
// 	//
// 	// -------------------------------------------------------------------
// 	//
//
// 	//
// 	// calculate difference in horizontal direction
// 	//
// 	// left boundary
// 	w1 = r+1;
//
// 	pJ_begin = J;
// 	pT2_begin = T + r*sy;
//
// 	//Concurrency::parallel_for( size_t(0), size_t(w1), [&](size_t ix)
// 	for ( size_t ix(0); ix < w1; ix++ )
// 	{
// 		size_t off(ix*sy);
// 		im_type
// 				*pJ(pJ_begin+off),
// 				*pT2(pT2_begin+off);
//
// 		for ( pJ_end = pJ+sy; pJ<pJ_end; pJ++, pT2++ )
// 		{	*pJ = *pT2;
// 		}
// 	}
// 	//});
//
//
// 	// normal filtering
// 	w2 = sx-2*r-1;
//
// 	pJ_begin = J + (r+1)*sy;
// 	pT1_begin = T;
// 	pT2_begin = T + (2*r+1)*sy;
//
// 	//Concurrency::parallel_for( size_t(0), size_t(w2), [&](size_t ix)
// 	for ( size_t ix(0); ix < w2; ix++ )
// 	{
// 		size_t off(ix*sy);
// 		im_type
// 			*pJ(pJ_begin+off),
// 			*pT1(pT1_begin+off),
// 			*pT2(pT2_begin+off);
//
// 		for ( pJ_end=pJ+sy; pJ<pJ_end; pJ++, pT1++, pT2++ )
// 		{	*pJ = *pT2 - *pT1;
// 		}
// 	}
// 	//});
//
// 	// right boundary
// 	w3 = r;
//
// 	pJ_begin = J + (sx-r)*sy;
// 	pT1_begin = T + (sx-2*r-1)*sy;
// 	pT2_begin = T + (sx-1)*sy;
//
// 	//Concurrency::parallel_for( size_t(0), size_t(w3), [&](size_t ix)
// 	for ( size_t ix(0); ix < w3; ix++ )
// 	{
// 		size_t off(ix*sy);
// 		im_type
// 			*pJ(pJ_begin+off),
// 			*pT1(pT1_begin+off),
// 			*pT2(pT2_begin);
//
// 		for ( pJ_end=pJ+sy; pJ<pJ_end; pJ++, pT1++, pT2++ )
// 		{	*pJ = *pT2 - *pT1;
// 		}
// 	}
// 	//});
// }
//
// #else

//
// Process in vertical direction
//
inline void box_filtering_hdir( size_t sy, size_t sx, im_type *J, const im_type *pI, im_type *T, size_t r ) {
	im_type *pJ, *pJ_end, *pT(T);
	im_type *pT1, *pT2, *pT2_begin;
	
	size_t w1, w2, w3;
	
	//
	// calculate cumulative sum
	//
	im_type *buff = new im_type [sy], *pbuff, *pbuff_end;
	std::fill( buff, buff+sy, 0 );
	
	for ( size_t ix = 0; ix < sx; ix++ ) {
		for ( im_type *pbuff(buff), *pbuff_end(buff+sy);  pbuff<pbuff_end;  pbuff++, pI++, pT++ ) {
			*pT = ( *pbuff += *pI );
		}
	}
	
	delete [] buff;
	
	
	//
	// -------------------------------------------------------------------
	//
	
	//
	// calculate difference in horizontal direction
	//
	// left boundary
	w1 = r+1;
	
	pJ = J;
	pT2 = T + (r*sy);
	
	for ( size_t ix = 0; ix < w1; ix++ ) {
		for ( pJ_end = pJ+sy; pJ<pJ_end; pJ++, pT2++ )
		{	*pJ = *pT2;
		}
	}
	
	
	// normal filtering
	w2 = sx-2*r-1;
	
	pJ = J + ((r+1)*sy);
	pT1 = T;
	pT2 = T + ((2*r+1)*sy);
	
	for ( size_t ix = 0; ix < w2; ix++ ) {
		for ( pJ_end=pJ+sy; pJ<pJ_end; pJ++, pT1++, pT2++ )
		{	*pJ = *pT2 - *pT1;
		}
	}
	
	
	// right boundary
	w3 = r;
	
	pJ = J + ((sx-r)*sy);
	pT1 = T + ((sx-2*r-1)*sy);
	pT2_begin = T + ((sx-1)*sy);
	
	for ( size_t ix = 0; ix < w3; ix++ ) {
		pT2 = pT2_begin;
		
		for ( pJ_end=pJ+sy; pJ<pJ_end; pJ++, pT1++, pT2++ )
		{	*pJ = *pT2 - *pT1;
		}
	}
}

// #endif

