//
// [V1, V2, V3, D] = imeig_power_c(
//     C11, C12, C13, C21, C22, C23, C31, C32, C33, 
//     U1, U2, U3,
//     nite, tole );
//
// This function calculates maximum eigen value D and vector V = [v1, v2, v3]
// of 3x3 matrix C = [c11, c12, c13; c21, c22, c23; c31, c32, c33]
// at each pixel of an image.
//
// Note that C11, C12, etc. are element images,
// that is, i'th pixel's Ci consists of Ci = [C11(i), C12(i), ... C33(i)].
//
//
// Input arguments
//
// Cs are covariance element images.
// Us are initial eigen vector images. User can also set null as [].
//
// nite is the number of iterations for power iteration (default nite=20).
// tole is the tolerance of convergence (default tole=0.00001).
//
//
// Output arguments
//
// Vs are eigen vector images.
// D is eigen value image.
//

#include <mex.h>
#include <cstdlib>
#include <algorithm>
#include <cmath>
// #include <omp.h>

#ifdef _MSC_VER
	#define __PARALLEL__
	#include <ppl.h>
#endif


//
// Prototype functions
//
void eig_power( double *V, double *D, const double *C, const double *U, double *buff );
void mtimes_mv( double *y, const double *A, const double *x, size_t m, size_t n );
double dot_vv( const double *a, const double *b, size_t m );
void mtimes_sv( double *a, const double s, size_t m );
double norm_v( const double *v, size_t m );


void find_max_eigpair( double *V1, double *V2, double *V3, double *D,
		const double *A11, const double *A12, const double *A13,
		const double *A21, const double *A22, const double *A23,
		const double *A31, const double *A32, const double *A33,
		const double *U1, const double *U2, const double *U3,
		const size_t npix, const size_t nite, const double tole );

void find_max_eigpair_multi( double *V1, double *V2, double *V3, double *D,
		const double *A11, const double *A12, const double *A13,
		const double *A21, const double *A22, const double *A23,
		const double *A31, const double *A32, const double *A33,
		const double *U1, const double *U2, const double *U3,
		const size_t sy, const size_t sx, const size_t nite, const double tole );


void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )
{
	//
	// ignore detail check
	//
	
	// get image size
	size_t ndim = mxGetNumberOfDimensions( prhs[0] );
	const mwSize *dims = mxGetDimensions( prhs[0] );
	size_t
			sy = dims[0],
			sx = dims[1],
			npix = sy*sx;
	
	double *A11, *A12, *A13, *A21, *A22, *A23, *A31, *A32, *A33;
	double *U1, *U2, *U3;
	size_t nite;
	double tole;
	
	A11 = (double*)mxGetData( prhs[0] );
	A12 = (double*)mxGetData( prhs[1] );
	A13 = (double*)mxGetData( prhs[2] );
	A21 = (double*)mxGetData( prhs[3] );
	A22 = (double*)mxGetData( prhs[4] );
	A23 = (double*)mxGetData( prhs[5] );
	A31 = (double*)mxGetData( prhs[6] );
	A32 = (double*)mxGetData( prhs[7] );
	A33 = (double*)mxGetData( prhs[8] );
	
	U1 = (double*)mxGetData( prhs[9] );
	U2 = (double*)mxGetData( prhs[10] );
	U3 = (double*)mxGetData( prhs[11] );
	
	nite = mxGetScalar( prhs[12] );
	tole = mxGetScalar( prhs[13] );

	//
	// output
	//
	double *V1, *V2, *V3, *D;
	
	plhs[0] = mxCreateNumericMatrix( sy, sx, mxDOUBLE_CLASS, mxREAL );
	plhs[1] = mxCreateNumericMatrix( sy, sx, mxDOUBLE_CLASS, mxREAL );
	plhs[2] = mxCreateNumericMatrix( sy, sx, mxDOUBLE_CLASS, mxREAL );
	plhs[3] = mxCreateNumericMatrix( sy, sx, mxDOUBLE_CLASS, mxREAL );
	
	V1 = (double*)mxGetData( plhs[0] );
	V2 = (double*)mxGetData( plhs[1] );
	V3 = (double*)mxGetData( plhs[2] );	
	D = (double*)mxGetData( plhs[3] );

	//
	// calculate the maximum eigen value and corresponding vector at each pixel
	//
// 	find_max_eigpair( V1, V2, V3, D,
// 			A11, A12, A13,
// 			A21, A22, A23,
// 			A31, A32, A33,
// 			U1, U2, U3,
// 			npix, nite, tole );
	
	find_max_eigpair_multi( V1, V2, V3, D,
			A11, A12, A13,
			A21, A22, A23,
			A31, A32, A33,
			U1, U2, U3,
			sy, sx, nite, tole );

}

//
// ------------------------------------------------------------------------
//

// global buffer
const size_t g_dim = 3;
double g_tole;
size_t g_nite;

//
// ------------------------------------------------------------------------
//

void find_max_eigpair( double *V1, double *V2, double *V3, double *D,
		const double *A11, const double *A12, const double *A13,
		const double *A21, const double *A22, const double *A23,
		const double *A31, const double *A32, const double *A33,
		const double *U1, const double *U2, const double *U3,
		const size_t npix, size_t nite=20, double tole=0.00001 )
{

	size_t idx;
	size_t d = 3;
	
	double unit = 1.0 / std::sqrt( (double)d );
	double Vi[3], Di[1], Ci[9], Ui[3], buff[6];	
	
	//
	// set global parameter
	//
	g_tole = tole;
	g_nite = nite;

	if ( U1==0 || U2==0 || U3==0 )
	{
		Ui[0] = unit;
		Ui[1] = unit;
		Ui[2] = unit;
		
		for ( idx = 0; idx < npix; ++idx )
		{
			Ci[0] = A11[idx];
			Ci[1] = A21[idx];
			Ci[2] = A31[idx];
			Ci[3] = A12[idx];
			Ci[4] = A22[idx];
			Ci[5] = A32[idx];
			Ci[6] = A13[idx];
			Ci[7] = A23[idx];
			Ci[8] = A33[idx];
			
			eig_power( Vi, Di, Ci, Ui, buff );
			
			// eigenvector is used as an initial eigen vector of the next pixel
			Ui[0] = V1[idx] = Vi[0];
			Ui[1] = V2[idx] = Vi[1];
			Ui[2] = V3[idx] = Vi[2];
			D[idx] = Di[0];
			
// 			// without eigenvector propagation
// 			V1[idx] = Vi[0];
// 			V2[idx] = Vi[1];
// 			V3[idx] = Vi[2];
// 			D[idx] = Di[0];
// 			
// 			Ui[0] = unit;
// 			Ui[1] = unit;
// 			Ui[2] = unit;
			
		}
	}
	else
	{
		for ( idx = 0; idx < npix; ++idx )
		{
			Ci[0] = A11[idx];
			Ci[1] = A21[idx];
			Ci[2] = A31[idx];
			Ci[3] = A12[idx];
			Ci[4] = A22[idx];
			Ci[5] = A32[idx];
			Ci[6] = A13[idx];
			Ci[7] = A23[idx];
			Ci[8] = A33[idx];

			Ui[0] = U1[idx];
			Ui[1] = U2[idx];
			Ui[2] = U3[idx];
			
			eig_power( Vi, Di, Ci, Ui, buff );
			
			V1[idx] = Vi[0];
			V2[idx] = Vi[1];
			V3[idx] = Vi[2];
			D[idx] = Di[0];
		}
	}
}

//
// ------------------------------------------------------------------------
//

void find_max_eigpair_multi( double *V1, double *V2, double *V3, double *D,
		const double *A11, const double *A12, const double *A13,
		const double *A21, const double *A22, const double *A23,
		const double *A31, const double *A32, const double *A33,
		const double *U1, const double *U2, const double *U3,
		const size_t sy, const size_t sx, size_t nite=20, double tole=0.00001 )
{
	size_t d = 3;
	double unit = 1.0 / std::sqrt( (double)d );

	//
	// set global parameter
	//
	g_tole = tole;
	g_nite = nite;

	if ( U1==0 || U2==0 || U3==0 )
	{
		#ifdef __PARALLEL__
		Concurrency::parallel_for( size_t(0), size_t(sx), [&](size_t ix) {
		#else		
		for (size_t ix = 0; ix < sx; ++ix ) {
		#endif
			double Vi[3], Di[1], Ci[9], Ui[3], buff[6];
		
			Ui[0] = unit;
			Ui[1] = unit;
			Ui[2] = unit;
				
			for ( size_t iy = 0; iy < sy; ++iy )
			{
				size_t idx = ix*sy + iy;
				
				Ci[0] = A11[idx];
				Ci[1] = A21[idx];
				Ci[2] = A31[idx];
				Ci[3] = A12[idx];
				Ci[4] = A22[idx];
				Ci[5] = A32[idx];
				Ci[6] = A13[idx];
				Ci[7] = A23[idx];
				Ci[8] = A33[idx];

				eig_power( Vi, Di, Ci, Ui, buff );

				// eigenvector is used as an initial eigen vector of the next pixel
				Ui[0] = V1[idx] = Vi[0];
				Ui[1] = V2[idx] = Vi[1];
				Ui[2] = V3[idx] = Vi[2];
				D[idx] = Di[0];

// 				// without eigenvector propagation
// 				V1[idx] = Vi[0];
// 				V2[idx] = Vi[1];
// 	 			V3[idx] = Vi[2];
// 				D[idx] = Di[0];
// 			
// 				Ui[0] = unit;
// 				Ui[1] = unit;
// 				Ui[2] = unit;
			}

		#ifdef __PARALLEL__
		});
		#else
		}
		#endif
	}
	else
	{
		#ifdef __PARALLEL__
		Concurrency::parallel_for( size_t(0), size_t(sx), [&](size_t ix) {
		#else		
		for (size_t ix = 0; ix < sx; ++ix ) {
		#endif
			double Vi[3], Di[1], Ci[9], Ui[3], buff[6];		
				
			for ( size_t iy = 0; iy < sy; ++iy )
			{
				size_t idx = ix*sy + iy;
				
				Ci[0] = A11[idx];
				Ci[1] = A21[idx];
				Ci[2] = A31[idx];
				Ci[3] = A12[idx];
				Ci[4] = A22[idx];
				Ci[5] = A32[idx];
				Ci[6] = A13[idx];
				Ci[7] = A23[idx];
				Ci[8] = A33[idx];

				Ui[0] = U1[idx];
				Ui[1] = U2[idx];
				Ui[2] = U3[idx];

				eig_power( Vi, Di, Ci, Ui, buff );

				V1[idx] = Vi[0];
				V2[idx] = Vi[1];
				V3[idx] = Vi[2];
				D[idx] = Di[0];
			}
		#ifdef __PARALLEL__
		});
		#else
		}
		#endif
	}
}

//
// ------------------------------------------------------------------------
//

// find eigen-vector and value of a matrix C by using power iteration
void eig_power( double *V, double *D, const double *C, const double *U, double *buff )
{
	double x_norm, error;
	size_t ite;
	
	double
			*x = buff,
			*x_ = buff + g_dim;
	
	// initial eigenvector
	std::copy( U, U+g_dim, x_ );
		
	for ( ite = 0; ite < g_nite; ++ite )
	{
		// update
		mtimes_mv( x, C, x_, g_dim, g_dim );

		// normalize
		x_norm = norm_v( x, g_dim );
		mtimes_sv( x, (1.0/x_norm), g_dim );

		// convergence check
		error = std::abs( dot_vv(x_, x, g_dim) - 1 );
		if ( error < g_tole )
		{
			break;
		}

		std::copy( x, x+g_dim, x_ );
	}

	std::copy( x, x+g_dim, V );
	D[0] = x_norm;
}

//
// ------------------------------------------------------------------------
//

// matrix times
void mtimes_mv( double *y, const double *A, const double *x, size_t m, size_t n )
{
	size_t i, j;
	const double *Aj;
	double xj;
	
	// calculate first columun of a matrix
	j = 0;
	Aj = A;
	xj = x[j];
	
	for ( i = 0; i < m; ++i )
	{
		y[i] = Aj[i] * xj;
	}
	
	// calculate other columns
	for ( j = 1; j < n; ++j )
	{
		Aj = A + j*m;
		xj = x[j];
	
		for ( i = 0; i < m; ++i )
		{
			y[i] += Aj[i] * xj;
		}
	}
}

//
// ------------------------------------------------------------------------
//

// dot( a[], b[] )
double dot_vv( const double *a, const double *b, size_t m )
{
	double sum = 0;
	
	for ( ; m > 0; --m )
	{
		sum += *a++ * *b++;
	}
	
	return sum;
}

//
// ------------------------------------------------------------------------
//

// norm( v[] )
double norm_v( const double *v, size_t m )
{
	double vi, sum = 0;
	
	for ( ; m > 0; --m )
	{
		vi = *v++;
		sum += vi*vi;
	}
	
	return std::sqrt( sum );
}

//
// ------------------------------------------------------------------------
//

// a[] = s * a[]
void mtimes_sv( double *a, const double s, size_t m )
{
	for ( ; m > 0; --m )
	{
		*a++ *= s;
	}
}


