function J = smooth_diffusion( I, hws, wstd, enhance, preserve, dt, nIter )
%
% edge preserving smoothing for the multi channel image
% using Tschumperle's isotropic diffusion.
%
%
% J = smooth_diffusion( I, hws, wstd, enhance, preserve, dt, nIter )
%
% Inputarguments
%
%   I : Multi channel image
%
%   smoothing parameters for gradient vectors
%   hws    : half window size of gaussian filter
%   wstd   : standard deviation
%
%   smoothing parameters
%   enhance  : preservation degree of edges (default [0.1 ~ 2])
%   preserve : preservation degree of original image (default [0.3 ~ 0.5])
%   dt       : time step
%	nIter    : number of iterations
%
%
% Output arguments
%
%	J : smoothed image
%
%
% REFERENCE
%
% D. Tschumperle, R. Deriche.
% Vector-valued image regularization with PDE's :
% A Common framework for different applications
% CVPR'2003
%

scale = enhance * 255^2;

global nColor;
[sy, sx, nColor] = size( I );
J = I;

% derivation operator mask
global Hx Hy Hxx Hxy Hyy;
Hx = 0.5 * [ -1, 0, 1 ];
Hy = Hx';

Hxx = [ 1 -2 1 ];
Hyy = Hxx';
Hxy = 0.25 * [ 1 0 -1; 0 0 0; -1 0 1 ];


% smoothing operator mask
global G_sigma;
G_sigma = fspecial( 'gaussian', 2*hws+1, wstd );



for iItr = 1:nIter
	
	% gradient of an image
	[ Ix Iy ] = gradient_( J );
	
	% tensor of gradients
	[ g11, g12, g21, g22 ] = structure_tensor( Ix, Iy );
	
	% Calculate the direction and the magnitude of gradients
	[ r_min, r_max, ux, uy, vx, vy ] = gradv( g11, g12, g21, g22 );
	
	r_min = scale * r_min;
	r_max = scale * r_max;
	
	% set parameters for the diffusion tensor
	r2 = 1 ./ ( 1 + r_min + r_max );
	r1 = sqrt( r2 );
	
	[ d11, d12, d21, d22 ] = diffusion_tensor( r1, r2, ux, uy, vx, vy );
	
	% calculate trace( DH )
	[ Ixx, Ixy, Iyx, Iyy ] = hessian_( J );
	DH = ( d11 .* Ixx + d12 .* Ixy ) + ( d21 .* Iyx + d22 .* Iyy );
	
	% orthonormal constraint
	L = preserve * ( J - I ) - (1-preserve)*DH;
	
	% update
	J = J + dt * (-L);
	
end

end

%
% -------------------------------------------------------------------------
%

function [ Ix Iy ] = gradient_( I )
global Hx Hy;
Ix = imfilter( I, Hx, 'replicate' );
Iy = imfilter( I, Hy, 'replicate' );
end

%
% -------------------------------------------------------------------------
%

function [ Ixx, Ixy, Iyx, Iyy ] = hessian_( I )
global Hxx Hxy Hyy;
Ixx = imfilter( I, Hxx, 'replicate' );
Ixy = imfilter( I, Hxy, 'replicate' );
Iyx = Ixy;
Iyy = imfilter( I, Hyy, 'replicate' );
end

%
% -------------------------------------------------------------------------
%

function 	[ g11, g12, g21, g22 ] = structure_tensor( Ix, Iy )
global G_sigma;
g11 = sum( Ix.*Ix, 3 );
g12 = sum( Ix.*Iy, 3 );
g22 = sum( Iy.*Iy, 3 );

g11 = imfilter( g11, G_sigma, 'replicate' );
g12 = imfilter( g12, G_sigma, 'replicate' );
g21 = g12;
g22 = imfilter( g22, G_sigma, 'replicate' );
end

%
% -------------------------------------------------------------------------
%

function [ d11, d12, d21, d22 ] = diffusion_tensor( r1, r2, ux, uy, vx, vy )
global nColor;
d11 = r1 .* ux.*ux + r2 .* vx.*vx;
d12 = r1 .* ux.*uy + r2 .* vx.*vy;
d22 = r1 .* uy.*uy + r2 .* vy.*vy;

d11 = repmat( d11, [ 1 1 nColor ] );
d12 = repmat( d12, [ 1 1 nColor ] );
d21 = d12;
d22 = repmat( d22, [ 1 1 nColor ] );
end

%
% -------------------------------------------------------------------------
%

function [ r_min, r_max, ux, uy, vx, vy ]...
	= gradv( g11, g12, g21, g22 )
%
% Calculate the direction and the magnitude of gradients
% by using eigen-value analysis
%

a = g11 + g22;
b = g11 - g22;
c = g12 .* g21;
d = sqrt( b.^2 + 4 * c );

r_min = 0.5 * ( a - d );
r_max = 0.5 * ( a + d );

r_min = max( r_min, 0 );
r_max = max( r_max, 0 );

t_max = atan2( r_max - g11, g12 );
vx = cos( t_max );
vy = sin( t_max );

ux = -vy;
uy =  vx;

end
