function J = smooth_nonlocal( I, para )
%
% SMOOTH_NONLOCAL is smoothing by non-local mean [1].
% In addition, for the purpose of acceleration, the integral image filtering [2]
% and the local filtering algorithm based on [3] are used.
%
%
% J = smooth_nonlocal( I, para )
%
% arguments
%
% - input
%
%   - I : input image
%
%   - para : parameters
%
%     para.w_radius = 5 is radius of the filter window.
%
%     para.k_radius = 5 is radius of the kernel.
%
%     para.std_range = 0.08 is standard deviation for similarity measure
%     and controles the degree of smoothness.
%
%
% - output
%
%   - J : output smoothed image
%
%
% REFERENCE
%
% [1] A. Buades, B. Coll, J. M. Morel,
% "A non-local algorithm for image denoising,"
% Proc. IEEE Int. Conf. on Computer Vision and Pattern Recognition, 2005,
% vol. 2, pp. 60-65.
%
% [2] F. C. Crow, "Summed-area tables for texture mapping,"
% in Proc. of the 11th Annual Conf. on Computer graphics and
% interactive techniques (SIGGRAPH), 1984, pp. 207?212.
%
% [3] L. Yuan, J. Sun, L. Quan, H. Y. Shum,
% "Progressive inter-scale and intra-scale non-blind image deconvolution,"
% ACM Trans. Graph., 2008, vol. 27, issue 3.
%
%


%
% parameters
%
% check specified parameters or set default parameters
para = parameter_setting( para );

rw = para.w_radius; % radius of the filter window
rk = para.k_radius; % radius of the similarity search
std_r = para.std_range;


[sy, sx, ncolor] = size( I );
 
%
% precalculation of a parameter of fixed value
%
npix_wnd = (2*rw+1)^2; % number of pixels in a window
param_beta = -0.5/std_r^2/npix_wnd;



% Output image
J = zeros( sy, sx, ncolor );
W = zeros( sy, sx );

%
% symmetric caluculation of 1st and 3rd guadrant
%
for dx = 0:rk
	%
	% symmetric odd padding
	%
	pad_x_i = [1+dx:-1:2, 1:sx];
	pad_x_j = [1:sx, sx-1:-1:sx-dx];
	
	pad_u_i = (1:sx);
	pad_u_j = (1+dx:sx+dx);
	
	for dy = 1:rk
		%
		% symmetric odd padding
		%
		pad_y_i = [1+dy:-1:2, 1:sy];
		pad_y_j = [1:sy, sy-1:-1:sy-dy];
		
		pad_v_i = (1:sy);
		pad_v_j = (1+dy:sy+dy);
	
		%
		% calculation of I(x) - I(y) 
		%
		
		% shifted versions of an image
		I_shift_i = I( pad_y_i, pad_x_i, : );
		I_shift_j = I( pad_y_j, pad_x_j, : );
		
		% square of difference
		D = I_shift_i - I_shift_j;
		D = D .* D;
		
		% sums toward the range dimension
		D_sum = sum( D, 3 );
		
		% take a sum of each local region
		% by using the integral image filtering
		D_sum = imboxfilter_c( D_sum, rw, rw );
		
		% weight conversion
		W_tmp = exp( param_beta * D_sum );

		%
		% convolution by moving averaging filtering
		%
		W_tmp_i = W_tmp( pad_v_i, pad_u_i );
		W_tmp_j = W_tmp( pad_v_j, pad_u_j );
		
		for i = 1:ncolor
			J(:,:,i) = J(:,:,i) +...
				W_tmp_i.*I_shift_i( pad_v_i, pad_u_i, i ) + ...
				W_tmp_j.*I_shift_j( pad_v_j, pad_u_j, i );
		end
		W = W + W_tmp_i + W_tmp_j;

	end
end


%
% symmetric caluculation of 2nd and 4th guadrant
%
for dx = -rk:-1
	
	%
	% symmetric odd padding (differ from 1st and 3rd quadrant)
	%
	pad_x_i = [1:sx, sx-1:-1:sx+dx];
	pad_x_j = [1-dx:-1:2, 1:sx];
	
	pad_u_i = (1-dx:sx-dx);
	pad_u_j = (1:sx);
	
	for dy = 0:rk
		%
		% symmetric odd padding
		%
		pad_y_i = [1+dy:-1:2, 1:sy];
		pad_y_j = [1:sy, sy-1:-1:sy-dy];
		
		pad_v_i = (1:sy);
		pad_v_j = (1+dy:sy+dy);
	
		%
		% calculation of I(x) - I(y) 
		%
		
		% shifted versions of an image
		I_shift_i = I( pad_y_i, pad_x_i, : );
		I_shift_j = I( pad_y_j, pad_x_j, : );
		
		% square of difference
		D = I_shift_i - I_shift_j;
		D = D .* D;
		
		% sums toward the range dimension
		D_sum = sum( D, 3 );
		
		% take a sum of each local region
		% by using the integral image filtering
		D_sum = imboxfilter_c( D_sum, rw, rw );
		
		% weight conversion
		W_tmp = exp( param_beta * D_sum );

		%
		% convolution by moving averaging filtering
		%
		W_tmp_i = W_tmp( pad_v_i, pad_u_i );
		W_tmp_j = W_tmp( pad_v_j, pad_u_j );
		
		for i = 1:ncolor
			J(:,:,i) = J(:,:,i) +...
				W_tmp_i.*I_shift_i( pad_v_i, pad_u_i, i ) + ...
				W_tmp_j.*I_shift_j( pad_v_j, pad_u_j, i );
		end
		W = W + W_tmp_i + W_tmp_j;

	end
end


% %
% %
% % non-symmetric guadrant calculation 
% %
% %
% for dx = -rk:rk
% 	%
% 	% symmetric odd padding
% 	%
% 	if dx >= 0
% 		pad_x = [1+dx:sx, sx-1:-1:sx-dx];
% 	else 
% 		pad_x = [-dx+1:-1:2, 1:sx+dx];
% 	end
% 
% 	for dy = -rk:rk
% 		%
% 		% symmetric odd padding
% 		%
% 		if dy >= 0
% 			pad_y = [1+dy:sy, sy-1:-1:sy-dy];
% 		else
% 			pad_y = [-dy+1:-1:2, 1:sy+dy];
% 		end
% 
% 		%
% 		% ignore center pixel
% 		%
% 		if ~( dx ~= 0 || dy ~= 0 ) % == (dx==0 && dy==0)
% 			continue;
% 		end
% 
% 		
% 		%
% 		% calculation of I(x) - I(x)Dx->y 
% 		%
% 		
% 		% a shifted version of an image
% 		I_shift = I( pad_y, pad_x, : );
% 		
% 		% square of difference
% 		D = I - I_shift;
% 		D = D .* D;
% 		
% 		% sums toward the range dimension
% 		D_sum = sum( D, 3 );
% 		
% 		% take a sum of each local region
% 		% by using the integral image filtering
% 		D_sum = imboxfilter_c( D_sum, rw, rw );
% 		
% 		% weight conversion
% 		W_tmp = exp( param_beta * D_sum );
% 
% 		%
% 		% convolution by moving averaging filtering
% 		%
% 		for i = 1:ncolor
% 			J(:,:,i) = J(:,:,i) + W_tmp.*I_shift(:,:,i);
% 		end
% 		W = W + W_tmp;
% 
% 	end
% end


%
% final normalization
%
W = 1./W;
for i = 1:ncolor
	J(:,:,i) = J(:,:,i) .* W;
end

end


%
%--------------------------------------------------------------------------
%

function para = parameter_setting( para )
%
% check specified parameters or set default parameters
%
if ~isfield( para, 'w_radius' )
	para.w_radius = 5;
end

if ~isfield( para, 'k_radius' )
	para.k_radius = 5;
end

if ~isfield( para, 'std_range' )
	para.std_range = 0.08;
end

end
