function J = smooth_wls_cg(I, lambda, alpha, nite )
%
% Edge-preserving smoothing based on the weighted least squares (WLS).
% This method corresponding a special case of the isotropic diffusion using
% weights come from log version of image
%
% J = smooth_wls_cg( I );
% J = smooth_wls_cg( I, lambda, alpha, L );
%
% Input arguments
%
% - I      : input image (gray scale)
% - lambda : Balances between the data term and the smoothness
%            term. Increasing lambda will produce smoother images.
%            Default value is 1.0
%       
% - alpha  : Gives a degree of control over the affinities by non-
%            lineary scaling the gradients. Increasing alpha will
%            result in sharper preserved edges. Default value: 1.2
%
% - nite   : the number of iteration
%
% Output arguments
%
% - J      : Smoothed image
%
%
% REFERENCE
% [1] Z. Farbman, R. Fattal, D. Lischinski, R. Szeliski,
%     "Edge-preserving decompositions for multi-scale tone and detail
%      manipulation,"
%

%
% This source code is modified version of original, modified by K.Shirai
% Last modified at 2011 Sep. 30
%

eps_ = 0.0001; % regularization

% parameter for CG solver
thresh_conv = 0.0004; % for CG iteration break
check_cnt = 10;       % check convergence by some iterations 

if (~exist('nite', 'var')) || isempty( nite ),
    nite = 15;
end

if(~exist('alpha', 'var')) || isempty( alpha ),
    alpha = 1.2;
end

if(~exist('lambda', 'var')) || isempty( lambda ),
    lambda = 1; % edge consideration
end

L = log(I+eps);


[sy, sx, ndim] = size(I);
npix = sy*sx;


% Compute affinities between adjacent pixels based on gradients of L
%
% Corresponding to eq.(6)
%

%
% Weight between 4 adjacent pixels
%
Dy = diff( L, 1, 1 );
Dx = diff( L, 1, 2 );

Dy = -lambda ./ ( mean( abs(Dy), 3 ).^alpha + eps_ );
Dx = -lambda ./ ( mean( abs(Dx), 3 ).^alpha + eps_ );

W_t = [Dy(1,:); Dy];   % top
W_b = [Dy; Dy(end,:)]; % bottom
W_l = [Dx(:,1), Dx];   % left
W_r = [Dx, Dx(:,end)]; % right
W_c = 1.0 - (W_t + W_b + W_l + W_r); % center


%
% solve equation by CG method
%
cnt = 0; % counter to check convergence

% set original image as the initial value
X = I;

Ap = calc_Lp( X, W_c, W_t, W_b, W_l, W_r ); % == Ax
b = I;

r = b - Ap; % residual
p = r; % conjugate vector

% CG iterations
for t = 1:nite

	Ap = calc_Lp( p, W_c, W_t, W_b, W_l, W_r );
	rr = dot( r(:), r(:) );
	
	alpha = rr / dot( p(:), Ap(:) );
	
	X = X + alpha * p;
 	r = r - alpha * Ap;

	% check convergence by every CG_CHECK_CNT iterations
	if cnt == check_cnt
		if ( norm(r(:))/(npix*ndim) < thresh_conv )
			break;
		end
		cnt = 0;
	end
	cnt = cnt + 1;
 	
	beta = dot( r(:), r(:) ) / rr;
	p = r + beta * p;
end

J = X;

end

%
%--------------------------------------------------------------------------
%

function Ap = calc_Lp( p, W_c, W_t, W_b, W_l, W_r )
%
% Calculation of Laplacian matrix
%

[sy,sx,ndim] = size(p);

Ap = zeros(sy,sx,ndim);

for i = 1:ndim
	Ap(:,:,i) ...
	   = W_c .* p(:,:,i) ... % center
	   + W_t .* [p(1,:,i); p(1:end-1,:,i)] ... % top
	   + W_b .* [p(2:end,:,i); p(end,:,i)] ... % bottom
	   + W_l .* [p(:,1,i), p(:,1:end-1,i)] ... % left
	   + W_r .* [p(:,2:end,i), p(:,end,i)];   % right
end

end






